from __future__ import annotations

from typing import Literal, Mapping, NewType

from typing_extensions import NotRequired, Unpack

from haitch._attrs import AttributeValue, GlobalAttrs
from haitch._void_elements._void_element import VoidElement

LinkElement = NewType("LinkElement", VoidElement)
"""A `<link>` element."""


def link(
    *,
    extra_attrs: Mapping[str, AttributeValue] = {},
    **attrs: Unpack[LinkAttrs],
) -> LinkElement:
    """Specifies relationships between the current document and an external resource.

    This element is most commonly used to link to stylesheets, but is also used to
    establish site icons (both "favicon" style icons and icons for the home screen
    and apps on mobile devices) among other things.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link>
    """
    el = VoidElement("link")(**attrs, **extra_attrs)
    return LinkElement(el)


class LinkAttrs(GlobalAttrs):
    """Attributes for the `<link>` element.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link>
    """

    href: str
    """This attribute specifies the URL of the linked resource.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#href>
    """

    rel: str
    """Defines the relationship between a linked resource and the current document.

    Valid values: <https://developer.mozilla.org/en-US/docs/Web/HTML/Attributes/rel>

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#rel>
    """

    as_: NotRequired[str]
    """Specifies the type of content being loaded.

    This attribute is required when `rel="preload"` has been set on the `<link>`
    element, optional when `rel="modulepreload"` has been set, and otherwise should
    not be used.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#as>
    """

    crossorigin: NotRequired[Literal["anonymous", "use-credentials"]]
    """Indicates whether CORS must be used when fetching the resource.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#crossorigin>
    """

    fetchpriority: NotRequired[Literal["high", "low", "auto"]]
    """Provides a hint of the relative priority to use when fetching a preloaded resource.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#fetchpriority>
    """

    hreflang: NotRequired[str]
    """Indicates the language of the linked resource.

    Note: use only if `href` attribute is present.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#hreflang>
    """

    integrity: NotRequired[str]
    """Contains inline-metadata to verify that the resource has not been manipulated.

    This can be a base64-encoded cryptographic hash of the resource (file) to be fetched.

    Note: An integrity value may contain multiple hashes separated by whitespace.
    A resource will be loaded if it matches one of those hashes.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#integrity>
    """

    media: NotRequired[str]
    """Specifies the media that the linked resource applies to.

    Its value must be a media type / media query. This attribute is mainly useful when
    linking to external stylesheets — it allows the user agent to pick the best adapted
    one for the device it runs on.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#media>
    """

    referrerpolicy: NotRequired[
        Literal[
            "no-referrer",
            "no-referrer-when-downgrade",
            "origin",
            "origin-when-cross-origin",
            "unsafe-url",
        ]
    ]
    """Indicates which referrer to use when fetching the resource.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#referrerpolicy>
    """

    sizes: NotRequired[str]
    """Defines the sizes of the icons for visual media contained in the resource.

    It must be present only if the `rel` contains a value of `icon` or a non-standard
    type such as Apple's `apple-touch-icon`. It may have the following values:

        - "any": can be scaled to any size, i.e. `image/svg+xml` format.
        - "<width>x<height>": each size is in pixels, i.e. "200x200".

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#sizes>
    """

    type_: NotRequired[str]
    """Defines the type of the content linked to.

     The value of the attribute should be a MIME type such as text/html, text/css, and
     so on. The common use of this attribute is to define the type of stylesheet being
     referenced (such as text/css), but given that CSS is the only stylesheet language
     used on the web, not only is it possible to omit the type attribute, but is
     actually now recommended practice. It is also used on rel="preload" link types,
     to make sure the browser only downloads file types that it supports.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#type>
    """



import haitch as H


def test_simple_h1_example():
    # Render known `h1` tag.
    h1 = H.h1("Hello, world")

    got = str(h1)
    want = "<h1>Hello, world</h1>"

    assert got == want


def test_custom_component():
    # Render custom `foo` tag (useful for web components).
    foo = H.foo("Hello, world")

    got = str(foo)
    want = "<foo>Hello, world</foo>"

    assert got == want


def test_customers_example():
    # Fetch customers from a data store.
    customers = [
        ("jane@aol.com", False, True),
        ("bob@example.com", True, False),
        ("mark@mail.org", True, False),
        ("karen@hr.org", False, False),
    ]

    # Build the DOM tree with attributes and children.
    dom = H.div(class_="container")(
        H.h1("Customers to contact:"),
        H.ul(id_="email-customer-list")(
            H.li(H.a(href=f"mailto:{email}")(email))
            for email, is_premium, is_new in customers
            if is_premium or is_new
        ),
    )

    got = str(dom)
    want = '<div class="container"><h1>Customers to contact:</h1><ul id="email-customer-list"><li><a href="mailto:jane@aol.com">jane@aol.com</a></li><li><a href="mailto:bob@example.com">bob@example.com</a></li><li><a href="mailto:mark@mail.org">mark@mail.org</a></li></ul></div>'
    # Pipe this into prettier for improved readability: $ echo '...' | prettier --parser html

    assert got == want



from __future__ import annotations

from typing import Mapping, NewType

from typing_extensions import Unpack

from haitch._attrs import AttributeValue, GlobalAttrs
from haitch._elements._element import Element
from haitch._typing import Child

H2Element = NewType("H2Element", Element)
"""A `<h2>` element."""


def h2(
    *children: Child,
    extra_attrs: Mapping[str, AttributeValue] = {},
    **attrs: Unpack[GlobalAttrs],
) -> H2Element:
    """Represents second level section heading.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/Heading_Elements>
    """
    el = Element("h2")(**attrs, **extra_attrs)(*children)
    return H2Element(el)



from __future__ import annotations

from typing import Mapping, NewType

from typing_extensions import NotRequired, Unpack

from haitch._attrs import AttributeValue, GlobalAttrs
from haitch._void_elements._void_element import VoidElement

BaseElement = NewType("BaseElement", VoidElement)
"""A `<base>` element."""


def base(
    *,
    extra_attrs: Mapping[str, AttributeValue] = {},
    **attrs: Unpack[BaseAttrs],
) -> BaseElement:
    """Specifies the base URL to use for all relative URLs in a document.

    Note: there can only be one `<base>` element in a document.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/base>
    """
    el = VoidElement("base")(**attrs, **extra_attrs)
    return BaseElement(el)


class BaseAttrs(GlobalAttrs):
    """Attributes for the `<base>` element.

    Warning: A `<base>` element must have an `href` attribute, a target attribute, or both.
    If at least one of these attributes are specified, the `<base>` element must come
    before other elements with attribute values that are URLs, such as a `<link>`'s `href`
    attribute.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/base>
    """

    href: NotRequired[str]
    """The base URL to be used throughout the document for relative URLs.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/base#href>
    """

    target: NotRequired[str]
    """A keyword or author-defined name of the default browsing context to display.

    The following keywords have special meaning: `_self`, `_blank`, `_parent`, `_top`.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/base#target>
    """



import haitch as H


def test_a_element():
    got = str(H.a(href="https://example.com")("Really cool site"))
    want = '<a href="https://example.com">Really cool site</a>'
    assert got == want


def test_body_element():
    got = str(H.body(onload='javascript:alert("document loaded")'))
    want = '<body onload="javascript:alert(&quot;document loaded&quot;)"></body>'
    assert got == want


def test_div_element():
    got = str(H.div("Generic container."))
    want = "<div>Generic container.</div>"
    assert got == want


def test_form_element():
    got = str(H.form(action="/add", method="post"))
    want = '<form action="/add" method="post"></form>'
    assert got == want


def test_h1_element():
    got = str(H.h1("Heading 1"))
    want = "<h1>Heading 1</h1>"
    assert got == want


def test_h2_element():
    got = str(H.h2("Heading 2"))
    want = "<h2>Heading 2</h2>"
    assert got == want


def test_h3_element():
    got = str(H.h3("Heading 3"))
    want = "<h3>Heading 3</h3>"
    assert got == want


def test_h4_element():
    got = str(H.h4("Heading 4"))
    want = "<h4>Heading 4</h4>"
    assert got == want


def test_h5_element():
    got = str(H.h5("Heading 5"))
    want = "<h5>Heading 5</h5>"
    assert got == want


def test_h6_element():
    got = str(H.h6("Heading 6"))
    want = "<h6>Heading 6</h6>"
    assert got == want


def test_head_element():
    got = str(H.head(H.link(href="main.css", rel="stylesheet")))
    want = '<head><link href="main.css" rel="stylesheet"/></head>'
    assert got == want


def test_html_element():
    got = str(H.html(lang="en"))
    want = '<!doctype html><html lang="en"></html>'
    assert got == want


def test_label_element():
    got = str(H.label(for_="username")("Username:"))
    want = '<label for="username">Username:</label>'
    assert got == want


def test_noscript_element():
    got = str(H.noscript("Only show when scripting disabled."))
    want = "<noscript>Only show when scripting disabled.</noscript>"
    assert got == want


def test_ol_element():
    ordered_list = H.ol(type_="i")(
        H.li("First"),
        H.li("Second"),
    )

    got = str(ordered_list)
    want = '<ol type="i"><li>First</li><li>Second</li></ol>'

    assert got == want


def test_p_element():
    got = str(H.p("This is a paragraph."))
    want = "<p>This is a paragraph.</p>"
    assert got == want


def test_pre_element():
    got = str(H.pre("Preformatted text."))
    want = "<pre>Preformatted text.</pre>"
    assert got == want


def test_script_element_src_script_file():
    got = str(H.script(async_=True, type_="module", src="async-script.js"))
    want = '<script async type="module" src="async-script.js"></script>'
    assert got == want


def test_script_element_inline_scripting():
    code = """
    const userInfo = JSON.parse(document.getElementById("data").text);
    console.log("User information: %o", userInfo);
"""

    got = str(H.script(code))
    want = """\
<script>
    const userInfo = JSON.parse(document.getElementById(&quot;data&quot;).text);
    console.log(&quot;User information: %o&quot;, userInfo);
</script>"""

    assert got == want


def test_span_element():
    got = str(H.span("Generic inline container."))
    want = "<span>Generic inline container.</span>"
    assert got == want


def test_style_element():
    styles = """
    p {
      color: blue;
      background-color: yellow;
    }
"""

    got = str(H.style(media="all and (max-width: 500px)")(styles))
    want = """\
<style media="all and (max-width: 500px)">
    p {
      color: blue;
      background-color: yellow;
    }
</style>"""

    assert got == want


def test_table_element():
    table = H.table(
        H.tr(
            H.th("Name"),
            H.th("Age"),
        ),
        H.tr(
            H.td("Mario"),
            H.td("33"),
        ),
    )

    got = str(table)
    want = "<table><tr><th>Name</th><th>Age</th></tr><tr><td>Mario</td><td>33</td></tr></table>"

    assert got == want


def test_ul_element():
    unordered_list = H.ul(
        H.li("First"),
        H.li("Second"),
    )

    got = str(unordered_list)
    want = "<ul><li>First</li><li>Second</li></ul>"

    assert got == want


def test_title_element():
    got = str(H.title("Hello, world!"))
    want = "<title>Hello, world!</title>"
    assert got == want



from __future__ import annotations

from typing import Mapping, NewType

from typing_extensions import Unpack

from haitch._attrs import AttributeValue, GlobalAttrs
from haitch._void_elements._void_element import VoidElement

WbrElement = NewType("WbrElement", VoidElement)
"""A `<wbr>` element."""


def wbr(
    *,
    extra_attrs: Mapping[str, AttributeValue] = {},
    **attrs: Unpack[GlobalAttrs],
) -> WbrElement:
    """Represents a word break opportunity.

    This is a position within text where the browser may optionally break a line, though
    its line-breaking rules would not otherwise create a break at that location

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/wbr>
    """
    el = VoidElement("wbr")(**attrs, **extra_attrs)
    return WbrElement(el)



import haitch as H


def test_fragment_element_does_not_render_its_parent_tag():
    dom = H.fragment(
        H.h1("Hello, world!"),
        H.p("Nice to meet you."),
    )

    got = str(dom)
    want = "<h1>Hello, world!</h1><p>Nice to meet you.</p>"

    assert got == want


def test_html5_component():
    html = H.html5(
        content=H.fragment(
            H.header(
                H.h1("Hello, reader."),
            ),
            H.main(
                H.section(
                    H.p("So why is this a cool page?"),
                    H.a(href="#")("Check the link for more info!"),
                ),
            ),
        ),
        page_title="Cool page",
        language_code="en",
        body_classes="container box",
        links=[
            H.link(href="main.css", rel="stylesheet"),
            H.link(href="custom.css", rel="stylesheet"),
        ],
        scripts=[
            H.script(src="main.js", defer=True),
        ],
    )

    got = str(html)
    want = '<!doctype html><html lang="en"><head><meta charset="utf-8"/><meta name="viewport" content="width=device-width, initial-scale=1"/><meta http-equiv="x-ua-compatible" content="ie=edge"/><title>Cool page</title><link href="main.css" rel="stylesheet"/><link href="custom.css" rel="stylesheet"/><script src="main.js" defer></script></head><body class="container box"><header><h1>Hello, reader.</h1></header><main><section><p>So why is this a cool page?</p><a href="#">Check the link for more info!</a></section></main></body></html>'
    # Pipe this into prettier for improved readability: $ echo '...' | prettier --parser html

    assert got == want


def test_html5_component_just_content():
    html = H.html5(content=H.main())

    got = str(html)
    want = '<!doctype html><html lang=""><head><meta charset="utf-8"/><meta name="viewport" content="width=device-width, initial-scale=1"/><meta http-equiv="x-ua-compatible" content="ie=edge"/></head><body><main></main></body></html>'
    # Pipe this into prettier for improved readability: $ echo '...' | prettier --parser html

    assert got == want



from __future__ import annotations

import html
from typing import Iterable, Mapping

from haitch._attrs import AttributeValue, serialize_attribute
from haitch._typing import Child, Html, SupportsHtml


class Element:
    """Lazily built HTML Element.

    An Element represents a Document Object Model (DOM) with optional
    attributes and children. Render the HTML to string by invoking the
    `__str__` method.
    """

    def __init__(self, tag: str) -> None:
        """Initialize element by providing a tag name, ie. "a", "div", etc."""
        self._tag = tag
        self._attrs: Mapping[str, AttributeValue] = {}
        self._children: Iterable[Child] = []

    def __call__(self, *children: Child, **attrs: AttributeValue) -> Element:
        """Add children and/or attributes to element.

        Provide attributes, children, or a combination of both:

        >>> import haitch as H
        >>> H.img(src="h.png", alt="Letter H")
        >>> H.h1("My heading")
        >>> H.h1(style="color: red;")("My heading")
        """
        if children:
            self._children = [*self._children, *children]

        if attrs:
            self._attrs = {**self._attrs, **attrs}

        return self

    def __str__(self) -> Html:
        """Renders the HTML element as a string."""
        return self._render()

    def _render(self) -> Html:
        attrs_ = "".join(serialize_attribute(k, v) for k, v in self._attrs.items())
        children_ = "".join(self._render_child(child) for child in self._children)

        if self._tag == "fragment":
            return Html(children_)

        html_str = "<%(tag)s%(attrs)s>%(children)s</%(tag)s>" % {
            "tag": self._tag,
            "attrs": attrs_,
            "children": children_,
        }

        if self._tag == "html":
            return Html("<!doctype html>" + html_str)

        return Html(html_str)

    def _render_child(self, child: Child) -> str:
        if child is None or child is False:
            return ""

        if isinstance(child, str):
            return html.escape(child)

        if isinstance(child, SupportsHtml):
            return child._render()

        if isinstance(child, Iterable):
            return "".join(str(nested_child) for nested_child in child)

        raise ValueError(f"Invalid child type: {type(child)}")



from __future__ import annotations

from typing import Mapping, NewType

from typing_extensions import Unpack

from haitch._attrs import AttributeValue, GlobalAttrs
from haitch._void_elements._void_element import VoidElement

HrElement = NewType("HrElement", VoidElement)
"""An `<hr>` element."""


def hr(
    *,
    extra_attrs: Mapping[str, AttributeValue] = {},
    **attrs: Unpack[GlobalAttrs],
) -> HrElement:
    """Represents a thematic break between paragraph-level elements.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/hr>
    """
    el = VoidElement("hr")(**attrs, **extra_attrs)
    return HrElement(el)



from __future__ import annotations

from typing import Literal, Mapping, NewType

from typing_extensions import NotRequired, Unpack

from haitch._attrs import AttributeValue, GlobalAttrs
from haitch._void_elements._void_element import VoidElement

MetaElement = NewType("MetaElement", VoidElement)
"""A `<meta>` element."""


def meta(
    *,
    extra_attrs: Mapping[str, AttributeValue] = {},
    **attrs: Unpack[MetaAttrs],
) -> MetaElement:
    """Represents metadata that cannot be represented by other meta-related elements.

    Other meta-related elements: `<base>`, `<link>`, `<script>`, `<style>`, `<title>`

    Note: the attribute `name` has a specific meaning for the `<meta>` element, and
    the `itemprop` attribute must not be set on the same `<meta>` element that has
    any existing `name`, `http-equiv` or `charset` attributes.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/meta>
    """
    el = VoidElement("meta")(**attrs, **extra_attrs)
    return MetaElement(el)


class MetaAttrs(GlobalAttrs):
    """Attributes for the `<meta>` element.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/meta>
    """

    charset: NotRequired[str]
    """Declares the document's character encoding.

    If the attribute is present, its value must be an ASCII case-insensitive match for
    the string "utf-8", because UTF-8 is the only valid encoding for HTML5 documents.
    `<meta>` elements which declare a character encoding must be located entirely
    within the first 1024 bytes of the document.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/meta#charset>
    """

    content: NotRequired[str]
    """Contains the valuef for the `http-equiv` or `name` attribute.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/meta#content>
    """

    http_equiv: NotRequired[
        Literal[
            "content-security-policy",
            "content-type",
            "default-style",
            "x-ua-compatible",
            "refresh",
        ]
    ]
    """Defines a pragma directive.

    The attribute is named `http-equiv(alent)` because all the allowed values are names
    of particular HTTP headers.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/meta#http-equiv>
    """

    name: NotRequired[str]
    """Provides metadata in terms of name-value pairs.

    Often used together with `content` where the `name` attribute gives the metadata
    name, and the `content` attributes gives the value.

    <https://developer.mozilla.org/en-US/docs/Web/HTML/Element/meta#name>
    """
